from sys import exc_info
from IPython import get_ipython
from typing import Type
import sys

_ipython = get_ipython()
_SHOW_TRACEBACK = True
if _ipython:
    _TRACEBACK_FUNCTION = _ipython.showtraceback


def _ipython_hide_traceback(*args, **kwargs):
    global _SHOW_TRACEBACK, _TRACEBACK_FUNCTION
    if _SHOW_TRACEBACK:
        return _TRACEBACK_FUNCTION(*args, **kwargs)
    else:
        _SHOW_TRACEBACK = True
        error_type, value, _ = exc_info()
        value.__cause__ = None  # suppress chained exceptions
        return _ipython._showtraceback(
            error_type,
            value,
            _ipython.InteractiveTB.get_exception_only(error_type, value),
        )


def _python_exception_handler(
    exception_type, exception, traceback, debug_hook=sys.excepthook
):
    global _SHOW_TRACEBACK
    if _SHOW_TRACEBACK:
        debug_hook(exception_type, exception, traceback)
    else:
        _SHOW_TRACEBACK = True
        print("%s: %s" % (exception_type.__name__, exception))


if _ipython:
    _ipython.showtraceback = _ipython_hide_traceback
else:
    sys.excepthook = _python_exception_handler


def raiser(
    message: str,
    type: Type[BaseException] = Exception,
    traceback: bool = False,
):
    """Raise an exception with or without the traceback.

    Args:
        message : The message of the exception.
        type : The type of the exception raised.
        traceback : The traceback is printed if False.

    """

    global _SHOW_TRACEBACK
    _SHOW_TRACEBACK = traceback
    raise type(message)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "ipoly"
copyright = "2022, Thomas Danguilhen"
author = "Thomas Danguilhen"
release = "0.1.3"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

import sys, os

sys.path.insert(0, os.path.abspath("../"))

extensions = [
    "sphinx.ext.napoleon",
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.autosummary",
    "sphinx.ext.githubpages",
]

autosummary_generate = True
autoclass_content = "both"

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"  # "sphinx_rtd_theme"
html_static_path = ["_static"]

import pathlib

# The readme that already exists
readme_path = pathlib.Path(__file__).parent.resolve().parent / "README.md"
# We copy a modified version here
readme_target = pathlib.Path(__file__).parent / "readme.md"

with readme_target.open("w") as outf:
    # Change the title to "Readme"
    outf.write(
        "\n".join(
            [
                "Readme",
                "======",
            ]
        )
    )
    lines = []
    for line in readme_path.read_text().split("\n"):
        if line.startswith("# "):
            # Skip title, because we now use "Readme"
            # Could also simply exclude first line for the same effect
            continue
        lines.append(line)
    outf.write("\n".join(lines))



import re
import os
import glob
from typing import Literal, Tuple, Iterable
import pyarrow.parquet as pq
import xlrd
import pandas as pd
import numpy as np
import pyarrow as pa
from colorama import Fore, Back, Style
from collections import Counter
import string
import copy
from typing import Optional, List

SILLY_DELIMITERS = frozenset(string.ascii_letters + string.digits + ".")


def caster(df: pd.DataFrame):
    string_cols = [col for col, col_type in df.dtypes.items() if col_type == "object"]
    if len(string_cols) > 0:
        mask = df.astype(str).apply(
            lambda x: x.str.match(r"(\d{1,4}[-/\\\. ]\d{1,2}[-/\\\. ]\d{2,4})+.*").any()
        )

        def excel_date(x):
            # Warning : Date in the French format
            x = pd.to_datetime(x, dayfirst=True, errors="ignore")
            x = x.apply(
                lambda y: xlrd.xldate.xldate_as_datetime(y, 0)
                if type(y) in (int, float)
                else y
            )
            return x.astype("datetime64[ns]")

        df.loc[:, mask] = df.loc[:, mask].apply(excel_date)
        del mask
    string_cols = [col for col, col_type in df.dtypes.items() if col_type == "object"]
    for col in string_cols:
        try:
            try:
                df[col] = df[col].str.split(",").str.join(".").values.astype(float)
            except AttributeError:
                df[col] = df[col].values.astype(float)
        except (ValueError, TypeError):
            pass

    df = df.apply(
        lambda col: col
        if not (
            col.dtype in [np.dtype("float64"), np.dtype("float32"), np.dtype("float16")]
        )
        or col.isna().any()
        or pd.Series((col.apply(np.int64) == col)).sum() != df.shape[0]
        else col.apply(np.int64),
        axis=0,
    )
    return df


def load(
    file: str | Iterable[str],
    sheet: int = 1,
    skiprows = None,
    on: str = "index",
    classic_data: bool = True,
    recursive: bool = True,
    has_title: bool = True,
    has_index: bool = True
):
    """Load files or folders for most used file types.

    Allowed file extensions are :
        - csv
        - xlsx
        - xls
        - txt
        - png
        - jpg
        - pkl
        - bmp
        - xlsm
        - json
        - parquet
        - wav

    Args:
        file: The file or folder name.
            If it is a folder, all supported file types will be
            loaded in a list.
        sheet:
        skiprows:
        on:
        classic_data:
        recursive:

    """

    if type(file) != str:
        return [
            load(elem, sheet, skiprows, on, classic_data, recursive) for elem in file
        ]
    split_path = file.split("/")
    if len(split_path) == 1:
        directory = "./"
    else:
        file = split_path[-1]
        directory = "/".join(split_path[:-1])
    del split_path
    file_format = file.split(".")[-1]
    if len(file_format) + 1 >= len(file):
        file_format = None
    if recursive:
        pathname = directory + "/**/" + file
    else:
        pathname = directory + "/" + file
    files = glob.glob(pathname, recursive=recursive)
    if len(files) > 1:
        print(
            "There are multiple files with '"
            + file
            + "' name in the directory/subdirectories"
        )  # TODO check multi extension
        raise Exception
    elif len(files) == 0:
        print("Warning : The file '" + file + "' wasn't found !")
        return pd.DataFrame()

    file = files[0].split(file)[0] + file
    match file_format:
        case "xlsx" | "xls" | "xlsm":
            excel = pd.ExcelFile(file)
            sheets = excel.sheet_names
            try:
                if type(sheet) is int:
                    extract = excel.parse(sheets[sheet - 1], skiprows=skiprows)
                else:
                    extract = excel.parse(sheet, skiprows=skiprows)
            except IndexError:
                print(
                    "There is no sheet number "
                    + str(sheet)
                    + ", please select a valid sheet."
                )
                raise IndexError
            extract.dropna(how="all", inplace=True)
            if len(
                [
                    True
                    for elem in extract.columns
                    if (type(elem) is str and "Unnamed" in elem)
                ]
            ) == len(extract.columns):
                extract, extract.columns = (
                    extract.drop(extract.head(1).index),
                    extract.head(1).values.tolist()[0],
                )
            if classic_data:
                if on != "index" and on is not None:
                    extract.dropna(subset=[on], inplace=True)
                extract = caster(extract)
            extract.set_index(extract.columns[0])
            extract.drop(extract.columns[0], axis=1, inplace=True)
        case "pkl":
            if not os.path.isfile(file):
                print("The specified pickle file doesn't exist !")
            extract = pd.read_pickle(file)
        case "csv":
            with open(file) as myfile:
                firstline = myfile.readline()
                delimiter = detect(firstline, default=";")
                myfile.close()
            print(delimiter)
            extract = pd.read_csv(file, delimiter=delimiter)
            extract.dropna(how="all", inplace=True) # Drop empty rows
            if not has_title:
                extract = extract.T.reset_index().T.reset_index(drop=True)
            if len(
                [
                    True
                    for elem in extract.columns
                    if (type(elem) is str and "Unnamed" in elem)
                ]
            ) == len(extract.columns):
                extract, extract.columns = (
                    extract.drop(extract.head(1).index),
                    extract.head(1).values.tolist()[0],
                )
            if classic_data:
                if on != "index":
                    extract.dropna(subset=[on], inplace=True) # Drop rows without label
                extract = caster(extract)
            if has_index:
                extract = extract.set_index(extract.columns[0]) # Set first column as index
        case "parquet":
            extract = pq.read_pandas(file).to_pandas()
        case "png" | "jpg":
            from cv2 import imread

            img = imread(file)
            if (img[:, :, 0] == img[:, :, 1]).all() and (
                img[:, :, 0] == img[:, :, 2]
            ).all():
                return img[:, :, 0]
            return img
        case "bmp":
            import imageio

            return imageio.v3.imread(file)
        case "json":
            import json

            with open(file) as user_file:
                file_contents = user_file.read()
            return json.loads(file_contents)
        case "txt":
            with open(file) as f:
                lines = f.read().splitlines()
            return lines
        case "wav":
            from librosa import load as librosa_load

            return librosa_load(file, sr=None)
        case None:  # Directory
            return [
                load(elem, sheet, skiprows, on, classic_data, recursive)
                for elem in glob.glob(file + "/*")
            ]
        case default:
            print(
                f"I don't handle this file format yet ({default}), came back in a decade."
            )
            raise Exception
    if classic_data:
        if on == "index":
            extract = extract[~extract.index.duplicated(keep="first")]
        elif (not (on in extract)) and (on != None):
            print(
                "There is no column name '"
                + on
                + "' in the sheet "
                + str(sheet)
                + " of the file '"
                + file
                + "' so I can't load this file? Try to change the 'on' parameter"
            )
            raise Exception
        else:
            extract.drop_duplicates(subset=on, keep="last", inplace=True)
        extract.replace("Null", np.nan, inplace=True)
    return extract


def save(
    object_to_save: pd.DataFrame | np.ndarray | dict,
    file: str,
    sheet="Data",
    keep_index=True,
):
    """Save different object types to different file types.

    Supported file types are:
        - pkl
        - parquet
        - json
        - xlsx
        - png

    Args:
        file: The file name.
        sheet: The sheet name the data is saved if saved in an excel.
        keep_index: Keep the indexes in the saved file.

    Raises:
        Exception: If the file can't be accessed or the file type is
        not supported.

    """

    match file.split(".")[-1]:
        case "xlsx":
            try:
                writer = pd.ExcelWriter(file)
            except PermissionError:
                print(
                    "I can't access the file '" + file + "', the "
                    "solution may be to close it and retry."
                )
                raise Exception
            try:
                object_to_save.to_excel(writer, sheet_name=sheet, index=keep_index)
            except IOError:
                object_to_save.to_excel(writer, sheet_name=sheet, index=True)
            # for column in object_to_save:
            #    column_length = max(object_to_save[column].astype(str).map(len).max(), len(column))
            #    col_idx = object_to_save.columns.get_loc(column)
            #    writer.sheets[sheet].set_column(col_idx, col_idx, column_length)
            writer._save()
        case "csv":
            object_to_save.to_csv("./" + file, index=keep_index)
        case "pkl":
            pd.to_pickle(object_to_save, "./" + file)
        case "parquet":
            caster(object_to_save)
            table = pa.Table.from_pandas(object_to_save, preserve_index=False)
            pq.write_table(table, file + ".parquet")
        case "png":
            from PIL import Image

            im = Image.fromarray(object_to_save)
            im.save(file)
        case "json":
            with open(file, "w") as outfile:
                outfile.write(object_to_save)
        case default:
            print(
                f"I don't handle this file format yet ({default}), came back in a decade."
            )
            raise Exception


def merge(
    df: pd.DataFrame,
    file: str | pd.DataFrame,
    sheet=1,
    on: str = "index",
    skiprows=None,
    how: str = "outer",
    save_file: bool = False,
):
    if not (type(df) is pd.DataFrame):
        print("The df parameter must be a DataFrame.")
        raise Exception
    if type(file) is pd.DataFrame:
        dataBase = file
    else:
        dataBase = load(
            file,
            sheet=sheet,
            skiprows=skiprows,
            on=on,
        )
    if not dataBase.empty:
        columns = list(dataBase.columns)
        if on != "index":
            try:
                columns.remove(on)
            except ValueError:
                print(
                    "You can't merge your data as there are no column '"
                    + on
                    + "' in your already loaded DataFrame."
                )
                raise Exception
        if df.empty:
            merge = dataBase.copy()
        elif on != "index":
            merge = dataBase.merge(df, how=how, on=on)
        else:
            merge = dataBase.merge(df, how=how, left_index=True, right_index=True)
        merge = merge.loc[:, ~merge.columns.duplicated()]
        col = list(merge.columns)
        if on != "index":
            col.remove(on)
        merge.dropna(how="all", subset=col, inplace=True)
        del dataBase
        drop_y = list(filter(lambda v: re.match(".*_y$", v), merge.columns))
        keep_x = list(filter(lambda v: re.match(".*_x$", v), merge.columns))
        keep = [name[:-2] for name in keep_x]
        for col_x, col_y in zip(keep_x, drop_y):
            merge[col_x] = merge[col_x].mask(pd.isnull, merge[col_y], errors="ignore")
        merge.drop(drop_y, axis=1, inplace=True)
        merge.rename(columns=dict(zip(keep_x, keep)), inplace=True)
        merge = merge.sort_index()
        merge.replace("Null", np.nan, inplace=True)
    else:
        merge = df
    if save_file:
        save(merge, file)
    else:
        return merge


def color(
    text: str,
    fg: Literal[
        "BLACK",
        "RED",
        "GREEN",
        "YELLOW",
        "BLUE",
        "MAGENTA",
        "CYAN",
        "WHITE",
        "LIGHTBLACK",
        "LIGHTRED",
        "LIGHTGREEN",
        "LIGHTYELLOW",
        "LIGHTBLUE",
        "LIGHTMAGENTA",
        "LIGHTCYAN",
        "LIGHTWHITE",
    ] = "RED",
    bg: Literal[
        "BLACK",
        "RED",
        "GREEN",
        "YELLOW",
        "BLUE",
        "MAGENTA",
        "CYAN",
        "WHITE",
        "LIGHTBLACK",
        "LIGHTRED",
        "LIGHTGREEN",
        "LIGHTYELLOW",
        "LIGHTBLUE",
        "LIGHTMAGENTA",
        "LIGHTCYAN",
        "LIGHTWHITE",
    ] = None,
) -> str:
    """Format the text to print it in color.

    Args:
        text: The input text to format.
        fg: The foreground color.
        bg: The background color.

    Examples:
        >>> "This is a " + color("blue text", "BLUE") + " !"
        'This is a \\x1b[34mblue text\\x1b[0m !'

        >>> print("This is a " + color("blue text", "BLUE") + " !")
        This is a \x1b[34mblue text\x1b[0m !

        >>> "This is a " + color("strange color", "GREEN", "WHITE") + " !"
        'This is a \\x1b[47m\\x1b[32mstrange color\\x1b[0m !'

    """

    if bg and ("LIGHT" in bg):
        bg += "_EX"
    if "LIGHT" in fg:
        fg += "_EX"
    colored_text = f"{Fore.__getattribute__(fg)}{text}{Style.RESET_ALL}"
    return colored_text if bg is None else Back.__getattribute__(bg) + colored_text


def _same(sequence):
    elements = iter(sequence)
    try:
        first = next(elements)
    except StopIteration:
        return True
    for el in elements:
        if el != first:
            return False
    return True


def detect(
    text: str,
    default: Optional[str] = None,
    whitelist: Optional[List[str]] = None,
    blacklist: Optional[List[str]] = SILLY_DELIMITERS,
) -> Optional[str]:
    r"""Detects the delimiter used in text formats such as CSV and its many counsins.

    >>> detect(r"looks|like|the vertical bar\nis|the|delimiter\n")
    '|'

    `detect_delimiter.detect()` looks at the text provided to try to
    find an uncommon delimiter, such as ` for whatever reason.

    >>> detect('looks\x10like\x10something stupid\nis\x10the\x10delimiter')
    '\x10'

    When `detect()` doesn't know, it returns `None`:

    >>> text = "not really any delimiters in here.\nthis is just text.\n"
    >>> detect(text)

    It's possible to provide a default, which will be used in that case:

    >>> detect(text, default=',')
    ','

    By default, it will prevent avoid checking alpha-numeric characters
    and the period/full stop character ("."). This can be adjusted via
    the `blacklist` parameter.

    If you believe that you know the delimiter, it's possible to provide
    a list of possible delimiters to check for via the `whitelist` parameter.
    If you don't provide a value, `[',', ';', ':', '|', '\t']` will be checked.
    """

    if whitelist:
        candidates = whitelist
    else:
        candidates = list(",;:|\t")

    sniffed_candidates = Counter()
    likely_candidates = []

    lines = []
    # todo: support streaming
    text_ = copy.copy(text)
    while len(lines) < 5:
        for line in text_.splitlines():
            lines.append(line)

    for c in candidates:
        fields_for_candidate = []

        for line in lines:
            for char in line:
                if char not in blacklist:
                    sniffed_candidates[char] += 1
            fields = line.split(c)
            n_fields = len(fields)

            # if the delimiter isn't present in the
            # first line, it won't be present in the others
            if n_fields == 1:
                break
            fields_for_candidate.append(n_fields)

        if not fields_for_candidate:
            continue

        if _same(fields_for_candidate):
            likely_candidates.append(c)

    # no delimiter found
    if not likely_candidates:
        if whitelist is None and sniffed_candidates:
            new_whitelist = [
                char for (char, _count) in sniffed_candidates.most_common()
            ]
            return detect(text, whitelist=new_whitelist) or default
        return default

    if default in likely_candidates:
        return default

    return likely_candidates[0]



from abc import ABC, abstractmethod
from typing import List
from ipoly.file_management import load, save


class Component(ABC):
    @abstractmethod
    def operation(self) -> str:
        pass


class Leaf(Component):
    def __init__(self, driver, scrap_object, file, backward_moves) -> None:
        self.driver = driver
        self.scrap_object = scrap_object
        self.file = file
        self.backward_moves = backward_moves

    def operation(self, df) -> str:
        df = self.scrap_object(self.driver, df, self.file)
        for backward_move in self.backward_moves:
            backward_move = _value_finder(self.driver, backward_move)
            click_element(self.driver, *backward_move)
        return df, 0


class Composite(Component):
    def __init__(
        self, forward_moves, driver, file, scrap_object, backward_moves
    ) -> None:
        self.forward_moves = forward_moves
        self.driver = driver
        self._children: List[Component] = []
        self.file = file
        self.scrap_object = scrap_object
        self.backward_moves = backward_moves

    def add(self, component: Component) -> None:
        self._children.append(component)

    def operation(self, df) -> str:
        from time import sleep

        sleep(2)
        forward_move = _value_finder(self.driver, self.forward_moves[0])
        elements = self.driver.find_elements(
            "xpath",
            "//"
            + forward_move[0]
            + "[@"
            + forward_move[1]
            + "='"
            + forward_move[2]
            + "']",
        ) + self.driver.find_elements(
            "xpath",
            "//"
            + forward_move[0]
            + "[@"
            + forward_move[1]
            + "='"
            + forward_move[2]
            + " ']",
        )
        if not self._children:
            for _ in elements:
                if len(self.forward_moves) == 1:
                    self.add(
                        Leaf(
                            self.driver,
                            self.scrap_object,
                            self.file,
                            self.backward_moves,
                        )
                    )
                else:
                    self.add(
                        Composite(
                            self.forward_moves[1:],
                            self.driver,
                            self.file,
                            self.scrap_object,
                            self.backward_moves,
                        )
                    )
        elements[len(elements) - len(self._children)].click()
        sleep(3)
        df, exploring = self._children[0].operation(df)
        if not exploring:
            self._children = self._children[1:]
        return df, len(self._children)


def _value_finder(driver, tag):
    from bs4 import BeautifulSoup

    soup = BeautifulSoup(driver.page_source, "html.parser")
    args, _ = unknown_attributes_finder((tag[0], {tag[1]: tag[2]}), {}, soup)
    return tag[0], tag[1], args[1][tag[1]]


def move(scrap_object, forward_moves, backward_moves):
    from bs4 import BeautifulSoup

    if not all(isinstance(el, list) for el in forward_moves):
        forward_moves = [forward_moves]
    if not all(isinstance(el, list) for el in backward_moves):
        backward_moves = [backward_moves]

    def move_function(driver, df, file):
        tree = Composite(forward_moves, driver, file, scrap_object, backward_moves)
        exploring = True
        while exploring:
            df, exploring = tree.operation(df)
        return df

    return move_function


def _scraping_driver(visible: bool = False, size=(1980, 1080)):
    from selenium import webdriver
    from selenium.webdriver.chrome.options import Options
    from selenium.webdriver.chrome.service import Service
    from webdriver_manager.chrome import ChromeDriverManager

    chrome_options = Options()
    if not visible:
        chrome_options.add_argument("--headless")
    chrome_options.add_argument(f"--window-size={size[0]},{size[1]}")
    return webdriver.Chrome(
        service=Service(ChromeDriverManager().install()), options=chrome_options
    )


def click_element(driver, categorie, classe, information):
    from selenium.common.exceptions import NoSuchElementException
    from time import sleep

    try:
        element = driver.find_element(
            "xpath", "//" + categorie + "[@" + classe + "='" + information + "']"
        )
    except NoSuchElementException:
        element = driver.find_element(
            "xpath", "//" + categorie + "[@" + classe + "='" + information + " ']"
        )
    element.click()
    sleep(3)


def unknown_attributes_finder(args, kwargs, soup):
    new_args = []
    new_kwargs = {}
    unknown_attr = None
    for arg in args:
        if type(arg) == dict:
            try:
                arg["class_"] = arg.pop("class")
            except KeyError:
                pass
            kwargs.update(arg)
        else:
            new_args.append(arg)
    for kwarg in kwargs.items():
        if type(kwarg[1]) == int:
            if unknown_attr:
                raise Exception("You can't look for multiple unknown attributes.")
            if kwarg[0] == "class_":
                unknown_attr = ("class", kwarg[1])
            else:
                unknown_attr = kwarg
        else:
            new_kwargs[kwarg[0]] = kwarg[1]
    if unknown_attr:
        search = soup.find_all(new_args, new_kwargs)
        search = [
            elem[unknown_attr[0]] for elem in search if elem.has_attr(unknown_attr[0])
        ]
        attr_values = []
        for elem in search:
            if " ".join(elem) not in attr_values:
                attr_values.append(" ".join(elem))
        if attr_values != []:
            new_args.append(
                {unknown_attr[0]: attr_values[min(unknown_attr[1], len(attr_values))]}
            )
    return new_args, new_kwargs


def find_all_text(*args, **kwargs):
    from re import compile as re_compile

    def func(soup):
        nonlocal args, kwargs
        args, kwargs = unknown_attributes_finder(args, kwargs, soup)
        args = [
            arg
            if type(arg) != dict
            else {k: re_compile(v + r" *") for k, v in arg.items()}
            for arg in args
        ]
        return [tag.text.strip() for tag in soup.find_all(*args, **kwargs)]

    return func


def scrap(*actions):
    from bs4 import BeautifulSoup
    from pandas import concat, DataFrame

    def scrap_function(driver, df, file):
        soup = BeautifulSoup(driver.page_source, "html.parser")
        scraped = dict()
        for action in actions:
            scraped[action[0]] = action[1](soup)
        df = concat([df, DataFrame(scraped)], ignore_index=True)
        save(df, file)
        return df

    return scrap_function


def scraper(url, file, scrap_object, visible: bool = False, size=(1980, 1080)):
    from time import sleep
    from pandas import DataFrame

    driver = _scraping_driver(visible, size)
    try:
        df = load(file)
    except FileNotFoundError:
        df = DataFrame(dtype="object")

    driver.get(url)
    sleep(3)
    scrap_object(driver, df, file)



from typing import Literal, Union

import pandas as pd
from pandas import DataFrame
from pylatex import (
    Document,
    Section,
    Subsection,
    Figure,
    Package,
    NoEscape,
    SubFigure,
    Command,
    Table,
)

import os

available_functions = Literal["section", "subsection", "image", "text", "table"]


def nothing(*_args, **_kwargs):
    pass


class LaTeX:
    """
    Class for creating pdf files easily from Python by writing LaTeX code and compiling it.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, *kwargs)
        self.doc = None
        self.chain = []

    def _image(
        self,
        func,
        image_path: Union[str, tuple[str]] = None,
        caption: Union[str, tuple[str]] = None,
        centering: bool = True,
    ):
        if type(image_path) is str:
            image_path = tuple(image_path)
            caption = tuple(caption)
        elif caption is None:
            caption = tuple(None for _ in range(0, len(image_path)))
        image_filenames = tuple(
            os.path.join(os.path.dirname(__file__), "..\\" + path)
            for path in image_path
        )

        def image_func(*args, **kwargs):
            with self.doc.create(Figure(position="H")) as _figure:
                if centering:
                    self.doc.append(Command("centering"))
                for image_filename, image_caption in zip(image_filenames, caption):
                    with self.doc.create(
                        SubFigure(width=NoEscape(rf"{1 / len(caption)}\linewidth"))
                    ) as image:
                        image.add_image(
                            image_filename, width=NoEscape(r"0.95\linewidth")
                        )
                        if image_caption:
                            image.add_caption(image_caption)
            func(*args, *kwargs)

        return image_func

    def _section(self, func, name: str):
        def section_func(*args, **kwargs):
            with self.doc.create(Section(name)):
                func(*args, *kwargs)

        return section_func

    def _subsection(self, func, name: str):
        def section_func(*args, **kwargs):
            with self.doc.create(Subsection(name)):
                func(*args, *kwargs)

        return section_func

    def _text(self, func, text: str):
        if text[-4:] == ".txt":
            with open(text, "r") as file:
                text = file.read()

        def text_func(*args, **kwargs):
            self.doc.append(text)
            func(*args, *kwargs)

        return text_func

    def _table(self, func, table: Union[str, DataFrame], name: str = None):
        if type(table) is str:
            try:
                if table[-4:] == ".csv":
                    table = pd.read_csv(table)
                elif table[-4:] == ".pkl":
                    table = pd.read_pickle(table)
                elif table[-5:] == ".xlsx":
                    # noinspection PyArgumentList
                    table = pd.read_excel(table, index_col=0)

            except PermissionError:
                self._raiser(
                    f"The file '{table}' is already open, please close it before trying again."
                )

        table = table.applymap(
            lambda x: "".join(
                c
                for c in x
                if c
                in ["\t", "\n", "\r"]
                + list(map(chr, range(224, 249)))
                + list(map(chr, range(32, 35)))
                + list(map(chr, range(38, 127)))
            )
            if type(x) == str
            else x
        )
        table = table.applymap(
            lambda x: x.replace("&", "\\&").replace("_", "\\_") if type(x) is str else x
        )

        def table_func(*args, **kwargs):
            with self.doc.create(Table(position="H")) as table_figure:
                if name:
                    table_figure.add_caption(name)
                table_figure.append(Command("centering"))
                table_figure.append(NoEscape(table.style.to_latex()))
            func(*args, *kwargs)

        return table_func

    def _chainer(self):
        self.doc = Document(inputenc="latin1")
        self.doc.packages.append(
            Package(
                "geometry",
                options=["tmargin=1cm", "lmargin=1cm", "rmargin=1cm", "bmargin=2cm"],
            )
        )
        self.doc.packages.append(Package("float"))
        self.doc.packages.append(Package("babel", options=["french"]))
        self.doc.append(r"\listoffigures")
        func_chain = nothing
        for func, args, kwargs in reversed(self.chain):
            func_chain = func(func_chain, *args, *kwargs)
        func_chain(nothing)

    def add(self, func: available_functions, *args, mandatory: bool = True, **kwargs):
        funcs = {
            "section": self._section,
            "subsection": self._subsection,
            "image": self._image,
            "text": self._text,
            "table": self._table,
        }
        if func == "image":
            if type(args[0]) == str:
                paths = tuple(args[0])
            else:
                paths = args[0]
            for path in paths:
                if not os.path.exists(
                    os.path.join(os.path.dirname(__file__), "..\\" + path)
                ):
                    if not mandatory:
                        return None
                    self._raiser(f"The path '{path}' doesn't correspond to an image.")
            if (
                type(args[0]) == tuple
                and len(args) > 1
                and len(args[0]) != len(args[1])
            ):
                if not mandatory:
                    return None
                self._raiser(
                    "The number of images must be equal to the number of captions."
                )
        if func == "table":
            if not os.path.exists(
                os.path.join(os.path.dirname(__file__), "..\\" + args[0])
            ):
                if not mandatory:
                    return None
                self._raiser(
                    f"The provided path '{args[0]}' for the table is incorrect."
                )
            elif args[0][-4:] not in (".csv", ".pkl", "xlsx"):
                if not mandatory:
                    return None
                self._raiser(f"The extension of the file '{args[0]}' is not handled.")
        self.chain.append((funcs[func], args, kwargs))

    def generate_pdf(self, filepath: str = None, compiler="pdfLaTeX"):
        self._chainer()
        self.doc.generate_pdf(filepath, clean_tex=True, compiler=compiler)

    def generate_tex(self, filepath: str = None):
        self._chainer()
        self.doc.generate_tex(filepath)



import random
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from pandas import DataFrame
from typing import Iterable, Any, Tuple, Literal
from sklearn import metrics
from nptyping import NDArray, Int
import pandas as pd
from ipoly.file_management import caster
import tensorflow as tf


def set_seed(seed: int = 42) -> None:
    """Set the seed fot NumPy, TensorFlow and PyTorch.

    Args:
        seed: The seed value to set.

    """

    from tensorflow import random as tfr
    from torch.backends import cudnn
    from torch.cuda import manual_seed as cuda_manual_seed
    from torch import manual_seed
    from os import environ

    random.seed(seed)
    np.random.seed(seed)
    environ["PYTHONHASHSEED"] = str(seed)
    tfr.set_seed(seed)
    cuda_manual_seed(seed)
    manual_seed(seed)
    cudnn.deterministic = True
    cudnn.benchmark = False


def plot_history(history, keys: list[str] = None) -> None:
    """Plot the history of a Tensorflow training for each metrics.

    Args:
        history: The history to plot.
        keys: List of all metrics that will be plot.Plot all metrics
            if not specified.

    """

    if keys is None:
        keys = history.history.keys()
    for key in keys:
        plt.plot(history.history[key])
        plt.plot(history.history["val_accuracy"])
        plt.legend(["train", "val"], loc="upper left")
        plot("model accuracy", "epoch", "accuracy")


def plot_correlation_matrix(df: DataFrame) -> None:
    """Plot the correlation matrix of your DataFrame.

    Args:
        df: The input DataFrame.

    """

    corr = df.corr()
    # Generate a mask for the upper triangle
    mask = np.triu(np.ones_like(corr, dtype=bool))
    # Set up the matplotlib figure
    f, ax = plt.subplots(figsize=(11, 9))
    # Generate a custom diverging colormap
    cmap = sns.diverging_palette(230, 20, as_cmap=True)
    # Draw the heatmap with the mask and correct aspect ratio
    sns.heatmap(
        corr,
        mask=mask,
        cmap=cmap,
        vmax=0.3,
        center=0,
        square=True,
        linewidths=0.5,
        cbar_kws={"shrink": 0.5},
    )


def _confusion_matrix(confusion_matrix, axes, class_names, fontsize=14):

    df_cm = DataFrame(
        confusion_matrix,
        index=class_names,
        columns=class_names,
    )

    try:
        heatmap = sns.heatmap(df_cm, annot=True, fmt="d", cbar=False, ax=axes)
    except ValueError:
        raise ValueError("Confusion matrix values must be integers.")
    heatmap.yaxis.set_ticklabels(
        heatmap.yaxis.get_ticklabels(), rotation=0, ha="right", fontsize=fontsize
    )
    heatmap.xaxis.set_ticklabels(
        heatmap.xaxis.get_ticklabels(), rotation=45, ha="right", fontsize=fontsize
    )
    axes.set_ylabel("True label")
    axes.set_xlabel("Predicted label")


def plot_confusion_matrix(
    y_pred: NDArray[Any, Int], y_true: NDArray[Any, Int], labels: Iterable[str] = None
):
    """Plot the confusion matrix.

    It determines by itself if it has to plot a simple confusion matrix
    or multiple  ones in case of multilabel classification.

    Args:
        y_pred: Predicted logits.
        y_true: True labels.
        labels: Labels name. Need to be specified only if it is a
            multilabel classification.

    """

    fig, ax = plt.subplots(2, 2, figsize=(10, 7))
    if ((np.unique(y_true.sum(axis=1)) == 1).all()) or (len(y_true.shape) == 1):
        if len(y_true.shape) == 2:
            y_true = y_true.argmax(axis=1)
            y_pred = y_pred.argmax(axis=1)
        _confusion_matrix(
            metrics.confusion_matrix(y_true, y_pred), ax.flatten(), ["N", "Y"]
        )
    else:
        confusion_matrix = metrics.multilabel_confusion_matrix(y_true, y_pred)
        for axes, cfs_matrix, label in zip(ax.flatten(), confusion_matrix, labels):
            _confusion_matrix(cfs_matrix, axes, ["N", "Y"])
            axes.set_title("Confusion Matrix for the class - " + label)

        fig.tight_layout()
        plt.show()


def croper(image: np.array, margin: int = 18) -> np.array:
    """Crop the white areason the borders of the input image.

    Args:
        image: The input image.
        margin: The margin in pixels kept around the image.

    Returns:
        The cropped image.

    """

    if len(np.unique(image)) == 1:
        raise Exception("The image is composed of a single color.")
    if len(image.shape) == 3:
        image_sum = image.sum(axis=2) % 765
    else:
        image_sum = image == 0
    true_points = np.argwhere(image_sum)
    top_left = true_points.min(axis=0)
    bottom_right = true_points.max(axis=0)
    return image[
        max(0, top_left[0] - margin) : bottom_right[0] + 1 + margin,
        max(0, top_left[1] - margin) : bottom_right[1] + 1 + margin,
    ]


def get_weighted_loss(pos_weights, neg_weights, epsilon=1e-7):
    """Compute the loss function of the model.

    Args:
        pos_weights : Positive frequencies of weights.
        neg_weights : Negative frequencies of weights.
        epsilon : To not divide by 0.

    Returns:
        The loss classic function for the lost function.

    """

    from keras import backend as K
    from tensorflow import cast, float32

    def weighted_loss(y_true, y_pred):
        y_true, y_pred = cast(y_true, float32), cast(y_pred, float32)
        # initialize loss to zero
        loss = 0.0

        for i in range(len([pos_weights])):
            # for each class, add average weighted loss for that class
            loss += K.mean(
                -(
                    (pos_weights * y_true * K.log(y_pred + epsilon))
                    + (neg_weights * (1 - y_true) * K.log(1 - y_pred + epsilon))
                )
            )  # complete this line
        return loss

    return weighted_loss


def plot(
    title: str = None,
    xlabel: str = None,
    ylabel: str = None,
    figsize: Any = None,
    grid: bool = False,
    xticks: Tuple[int] = None,
    yticks: Tuple[int] = None,
):
    fig = plt.figure(figsize=figsize, dpi=100)
    fig.patch.set_facecolor("xkcd:white")
    plt.xticks(rotation=45, ha="right")

    plt.legend(bbox_to_anchor=(1.05, 1))
    if xticks:
        plt.yticks(np.arange(*xticks))
    if yticks:
        plt.yticks(np.arange(*yticks))
    plt.grid(grid)
    if ylabel:
        plt.ylabel(ylabel)
    if xlabel:
        plt.xlabel(xlabel)
    if title:
        plt.title(title)
    if ".png" in title:
        plt.savefig(title)
    else:
        plt.show()
    plt.close()


def path(path: str) -> str:
    from platform import system as ps

    if ps() == "Windows":
        return path.replace("/", "\\")
    return path.replace("\\", "/")


def interpolator(df: pd.DataFrame) -> pd.DataFrame:
    from scipy.interpolate import NearestNDInterpolator

    array = np.array(df)
    filled_array = array[~np.isnan(array).any(axis=1), :]
    for i in range(array.shape[1]):
        idxs = list(range(array.shape[1]))
        idxs.pop(i)
        my_interpolator = NearestNDInterpolator(
            filled_array[:, idxs], filled_array[:, i], rescale=True
        )
        array[:, i] = np.apply_along_axis(
            lambda row: my_interpolator(*row[idxs]) if np.isnan(row[i]) else row[i],
            1,
            array,
        )
    return caster(pd.DataFrame(array, columns=df.columns, index=df.index))


def prepare_table(
    df: pd.DataFrame,
    y: str | list[str] = "",
    correlation_threshold: float = 0.9,
    missing_rows_threshold: float = 0.9,
    missing_columns_threshold: float = 0.6,
    categories_ratio_threshold: float = 0.1,
    id_correlation_threshold: float = 0.04,
    verbose: bool = False,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Prepare the table to feed a ML model.

    Args:
        df : The DaaFrame to prepare for Machine Learning.
        y : Column(s) name(s) of target variable(s).

    """

    y = [y] if not y is list else y
    category_columns = df.select_dtypes(include=["category", object]).columns
    # Drop rows with NaNs in tye y columns
    if y != [""]:
        for col in y:
            df = df[df[col].notna()]
    # Drop the categorical columns with too much categories
    df = df.drop(
        [
            col
            for col in category_columns
            if df[col].nunique() / len(df) > categories_ratio_threshold
        ],
        axis=1,
    )
    # Convert categorical data to numerical ones
    df = pd.get_dummies(df)
    # Drop columns with not enough data
    df = df.loc[
        :,
        df.apply(
            lambda col: (1 - col.count() / len(df.index)) < missing_columns_threshold,
            axis=0,
        ),
    ]
    # Drop rows with not enough data
    df = df[
        df.apply(
            lambda row: (1 - row.count() / len(df.columns)) < missing_rows_threshold,
            axis=1,
        )
    ]
    correlate_couples = []
    corr = df.corr().abs()
    for col in corr:
        for index, val in corr[col].items():
            if val > correlation_threshold and (index != col):
                if (index, col) not in correlate_couples:
                    correlate_couples.append((col, index))
        if (df[col].nunique() == df.shape[0]) and (
            (corr[col].sum() - 1) / corr.shape[0] < id_correlation_threshold
        ):  # TODO améliorer
            # Drop ids columns (unique values with low correlations with other columns)
            if not col in y:
                df = df.drop(col, axis=1)
    for couple in correlate_couples:
        # Drop a column if highly correlated with another one
        if not any(elem in couple for elem in y):
            df = df.drop(couple[0], axis=1)
    if verbose:
        print("")
    return interpolator(df.drop(y, axis=1)), df[y]


def get_optimizer(
    type: Literal["Adam", "SGD", "RMSprop", "Adadelta", "Adamax", "Nadam", "Ftrl"],
    learning_rate,
):
    from tensorflow import keras

    match type:
        case "Adam":
            optimizer = keras.optimizers.Adam(learning_rate)
        case "SGD":
            optimizer = keras.optimizers.SGD(learning_rate)
        case "RMSprop":
            optimizer = keras.optimizers.RMSprop(learning_rate)
        case "Adadelta":
            optimizer = keras.optimizers.Adadelta(learning_rate)
        case "Adadelta":
            optimizer = keras.optimizers.Adagrad(learning_rate)
        case "Adadelta":
            optimizer = keras.optimizers.Adamax(learning_rate)
        case "Adadelta":
            optimizer = keras.optimizers.Nadam(learning_rate)
        case "Adadelta":
            optimizer = keras.optimizers.Ftrl(learning_rate)
    return optimizer


class SavePretrainedCallback(tf.keras.callbacks.Callback):
    # Hugging Face models have a save_pretrained() method that saves both the weights and the necessary
    # metadata to allow them to be loaded as a pretrained model in future. This is a simple Keras callback
    # that saves the model with this method after each epoch.
    def __init__(self, output_dir):
        super().__init__()
        self.output_dir = output_dir

    def on_epoch_end(self, _epoch, _logs=None):
        self.model.save_pretrained(self.output_dir)


def get_callbacks():
    """Get the main callbacks for a Tensorflow model."""

    import tensorflow_addons as tfa
    from datetime import datetime

    def scheduler(_, lr):
        return lr * tf.math.exp(-0.1)

    scheduler = tf.keras.callbacks.LearningRateScheduler(scheduler)
    logdir = "logs/image/" + datetime.now().strftime("%Y%m%d-%H%M%S")
    tensorboard_callback = tf.keras.callbacks.TensorBoard(
        log_dir=logdir, histogram_freq=1
    )
    return [
        tensorboard_callback,
        # tfa.callbacks.TQDMProgressBar(),
        tf.keras.callbacks.EarlyStopping(monitor="val_loss", patience=1),
        scheduler,
    ]


def compile_model(model, is_regression: bool, *args, **kwargs):
    """Compile a Tensorflow model.

    Args:
        model : The model architecture to compile.
        is_regression : Whenever the prediction task is a classification one or a regression one.
    """

    import tensorflow_addons as tfa

    optimizer = get_optimizer(*args, **kwargs)

    if is_regression:
        loss_fn = tf.keras.losses.MeanSquaredError()
        metrics = [tfa.metrics.F1Score(num_classes=2, average="micro")]
    else:
        loss_fn = tf.keras.losses.SparseCategoricalCrossentropy()
        metrics = ["accuracy"]
    model.compile(optimizer=optimizer, loss=loss_fn, metrics=metrics)
    return model


def load_transformers(
    overwrite_output_dir: bool,
    output_dir: str,
    num_labels: int,
    model_name_or_path,
    cache_dir,
    model_revision,
    use_auth_token: bool,
    tokenizer_name: str = None,
    config_name: str = None,
    config_changes: dict = None,
):
    """Load a transformer, a configuration and a tokenizer from 'transformers' library.

    Args:
        tokenizer_name : The name of the tokenizer from HuggingFace.
        config_name : The name of the configuration from HuggingFace.
    """

    from transformers import AutoConfig
    from transformers import AutoTokenizer
    from transformers import TFAutoModelForSequenceClassification
    from transformers.utils import CONFIG_NAME, TF2_WEIGHTS_NAME
    from os import listdir
    from os.path import isdir

    if not tokenizer_name:
        tokenizer_name = model_name_or_path
    if not config_name:
        config_name = model_name_or_path

    checkpoint = None
    if isdir(output_dir) and len(listdir(output_dir)) > 0 and not overwrite_output_dir:
        if (output_dir / CONFIG_NAME).is_file() and (
            output_dir / TF2_WEIGHTS_NAME
        ).is_file():
            checkpoint = output_dir
        else:
            raise ValueError(
                f"Output directory ({output_dir}) already exists and is not empty. "
                "Use --overwrite_output_dir to continue regardless."
            )

    if checkpoint is not None:
        config_path = output_dir
    elif config_name:
        config_path = config_name
    else:
        config_path = model_name_or_path
    if num_labels is not None:
        config = AutoConfig.from_pretrained(
            config_path,
            num_labels=num_labels,
            cache_dir=cache_dir,
            revision=model_revision,
            use_auth_token=True if use_auth_token else None,
        )
    else:
        config = AutoConfig.from_pretrained(
            config_path,
            cache_dir=cache_dir,
            revision=model_revision,
            use_auth_token=True if use_auth_token else None,
        )
    tokenizer = AutoTokenizer.from_pretrained(
        tokenizer_name if tokenizer_name else model_name_or_path,
        cache_dir=cache_dir,
        revision=model_revision,
        use_auth_token=True if use_auth_token else None,
    )
    if checkpoint is None:
        model_path = model_name_or_path
    else:
        model_path = checkpoint
    if config_changes:
        for key, value in config_changes.items():
            config.__setattr__(key, value)
    try:
        model = TFAutoModelForSequenceClassification.from_pretrained(
            model_path,
            config=config,
            cache_dir=cache_dir,
            revision=model_revision,
            use_auth_token=True if use_auth_token else None,
        )
    except OSError:
        model = TFAutoModelForSequenceClassification.from_pretrained(
            model_path,
            config=config,
            from_pt=True,
            cache_dir=cache_dir,
            revision=model_revision,
            use_auth_token=True if use_auth_token else None,
        )
    return model, tokenizer


def flatten(l):
    """Flatten a list.

    Args:
        l : The list to flatten.

    """

    return [item for sublist in l for item in sublist]


def compute_position(sentence: str | list, target: int, tokenizer) -> int:
    """Compute the first token's position of the targeted word in a sentence.

    Args:
        sentence : The sentence.
        target : The position of the targeted word in the sentence.
        tokenizer : The tokenizer to compute tokens.

    """

    if type(sentence) == str:
        sentence = sentence.split(" ")
    position = 0
    for i in range(target):
        word = sentence[i]
        if i == target:
            break
        position += len(tokenizer.encode(word)) - 2
    return position + 2


def train(model, train, val, epochs, batch_size):
    return model.fit(
        train[0],
        train[1],
        validation_data=val,
        epochs=epochs,
        batch_size=batch_size,
        callbacks=get_callbacks(),
    )


def evaluate(model, train, val, test):
    train_loss, train_metrics = model.evaluate(train[0], train[1])
    val_loss, val_metrics = model.evaluate(val[0], val[1])
    test_loss, test_metrics = model.evaluate(test[0], test[1])
    return [
        [train_loss, train_metrics],
        [val_loss, val_metrics],
        [test_loss, test_metrics],
    ]


def get_strategy(mixed_precision: bool = False, xla_accelerate: bool = False):
    """Get the Tensorflow strategy.

    Args:
        mixed_precision : Mixed precision is enabled if True.
        xla_accekerate : XLA acceleration is enabled if True.

    Returns:
        The distribute strategy if you are not using a TPU and the experimental distribute strategy if you are.

    """

    try:
        tpu = tf.distribute.cluster_resolver.TPUClusterResolver()
    except ValueError:
        tpu = None

    if tpu:
        tf.config.experimental_connect_to_cluster(tpu)
        tf.tpu.experimental.initialize_tpu_system(tpu)
        strategy = tf.distribute.experimental.TPUStrategy(tpu)
    else:
        strategy = tf.distribute.get_strategy()
    if mixed_precision:
        if tpu:
            policy = tf.keras.mixed_precision.experimental.Policy("mixed_bfloat16")
        else:
            policy = tf.keras.mixed_precision.experimental.Policy("mixed_float16")
        tf.keras.mixed_precision.set_policy(policy)
    if xla_accelerate:
        tf.config.optimizer.set_jit(True)
    return strategy



from typeguard.importhook import install_import_hook

with install_import_hook("ipoly"):
    from ipoly.file_management import *
    from ipoly.LaTeX import *
    from ipoly.communication import *
    from ipoly.logger import *
    from ipoly.traceback import *
    from ipoly.ml import *
    from ipoly.scraping import *



from pytest import raises
from sys import path

from ipoly import *


def test_load():
    assert load("test.txt") == ["First", "Second"]
    assert load("image1.jpg").shape == (220, 220, 3)
    assert load("image2.png").shape == (435, 358, 3)
    with raises(TypeError):
        load(1)
    assert load("test_folder")[0].sum() == 34703437
    assert len(load("test_folder")) == 2



import smtplib
from email.mime.application import MIMEApplication
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Union
import imaplib
from plyer import notification
import os


def send(
    sender: str,
    password: str,
    receivers: Union[str, list[str]],
    subject: str,
    content: str,
    pngfiles: list[str] = (),
):
    """Send an email to (a/some) receiver(s).

    Args:
        sender: The sender email. Gmail and Outlook emails are supported.
        password: The sender's password.
        receivers: List or single email of the receiver(s).
        subject: The object of the email.
        content: The main text part of the email.
        pngfiles: List of attached files.

    """
    msg = MIMEMultipart()
    msg["Subject"] = subject
    msg["From"] = sender
    msg["To"] = ", ".join(receivers if receivers is list else [receivers])
    msg.attach(MIMEText(content, "plain"))
    for file in pngfiles:
        extension = file.split(r".")[-1]
        with open(file, "rb") as fp:
            if extension == "png":
                attached_file = MIMEImage(fp.read())
                # Open the files in binary mode.  Let the MIMEImage class automatically
                # guess the specific image type.
            else:
                attached_file = MIMEApplication(fp.read(), Name=file)
        msg.attach(attached_file)

    send_port = 587

    # Send the email via our own SMTP server.
    host = "smtp.gmail.com" if "gmail" in sender else "smtp.office365.com"
    with smtplib.SMTP(host, send_port) as smtp:
        smtp.ehlo()
        smtp.starttls()
        smtp.login(sender, password)
        smtp.sendmail(sender, receivers, msg.as_string())
        smtp.quit()


def receive(receiver: str, password: str):
    # connect to the server and go to its inbox
    mail = imaplib.IMAP4_SSL("imap.gmail.com")
    mail.login(receiver, password)
    # we choose the inbox but you can select others
    mail.select("inbox")
    # we'll search using the ALL criteria to retrieve
    # every message inside the inbox
    # it will return with its status and a list of ids
    status, data = mail.search(None, "ALL")
    # the list returned is a list of bytes separated
    # by white spaces on this format: [b'1 2 3', b'4 5 6']
    # so, to separate it first we create an empty list
    mail_ids = []
    # then we go through the list splitting its blocks
    # of bytes and appending to the mail_ids list
    for block in data:
        # the split function called without parameter
        # transforms the text or bytes into a list using
        # as separator the white spaces:
        # b'1 2 3'.split() => [b'1', b'2', b'3']
        mail_ids += block.split()

    # now for every id we'll fetch the email
    # to extract its content
    for i in mail_ids:
        # the fetch function fetch the email given its id
        # and format that you want the message to be
        status, data = mail.fetch(i, "(RFC822)")

        # the content data at the '(RFC822)' format comes on
        # a list with a tuple with header, content, and the closing
        # byte b')'
        for response_part in data:
            # so if its a tuple...
            if isinstance(response_part, tuple):
                # we go for the content at its second element
                # skipping the header at the first and the closing
                # at the third
                message = email.message_from_bytes(response_part[1])

                # with the content we can extract the info about
                # who sent the message and its subject
                mail_from = message["from"]
                mail_subject = message["subject"]

                # then for the text we have a little more work to do
                # because it can be in plain text or multipart
                # if its not plain text we need to separate the message
                # from its annexes to get the text
                if message.is_multipart():
                    mail_content = ""

                    # on multipart we have the text message and
                    # another things like annex, and html version
                    # of the message, in that case we loop through
                    # the email payload
                    for part in message.get_payload():
                        # if the content type is text/plain
                        # we extract it
                        if part.get_content_type() == "text/plain":
                            mail_content += part.get_payload()
                else:
                    # if the message isn't multipart, just extract it
                    mail_content = message.get_payload()

                # and then let's show its result
                print(f"From: {mail_from}")
                print(f"Subject: {mail_subject}")
                print(f"Content: {mail_content}")


def say(message: str) -> None:
    """Make the computer say the message out loud.

    This function works on Linux/Windows and Mac platforms.

    Args:
        message: The message the computer says.
            For security, the characters used must be in this list:\n
            0123456789,;:.?!-_ÂÃÄÀÁÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿ

    """

    from string import ascii_letters

    if all(
        char
        in ascii_letters
        + "0123456789,;:.?!-_ÂÃÄÀÁÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿ"
        for char in message
    ):
        raise Exception(
            "This message will not be said because you used some symbols that may be used for doing some malicious injection :("
        )
    from platform import system as ps
    from os import system as oss

    match ps():
        case "Windows":
            from win32com.client import Dispatch

            speak = Dispatch("SAPI.SpVoice").Speak
            speak(message)
        case "Darwin":
            oss(f"say '{message}' &")
        case "Linux":
            oss(f"spd-say '{message}' &")
        case syst:
            raise RuntimeError("Operating System '%s' is not supported" % syst)


def notify(message: str, title: str = "Hey !") -> None:
    """Send a notification.

    This function works on Linux/Windows and Mac platforms and uses plyer in backend.

    Args:
        message: The message sent as a notification.
        title: The title of the notification. Defaults to 'Hey !'.

    """

    notification.notify(
        app_name="iPoly",
        app_icon=os.path.join(os.path.dirname(__file__), r"img\ipoly.ico"),
        title=title,
        message=message,
        ticker="iPoly ticker",
        timeout=15,
    )
